Require Import Coq.Lists.List
        Coq.Strings.String
        Coq.Arith.Arith.
Require Coq.Vectors.Vector.

(* Typeclasses for ensuring that a string is included
   in a list (i.e. a set of method names). This allows
   us to omit a default case (method not found) for method
   lookups. *)

Section BoundedIndex.

  Context {A : Type}.

  Class IndexBound n
        (a : A)
        (Bound : Vector.t A n)
    :=
      { ibound : Fin.t n;
        boundi : Vector.nth Bound ibound = a
      }.

  Global Arguments ibound {n a Bound} IndexBound.
  Global Arguments IndexBound {n} a Bound.

  Record BoundedIndex
         (n : nat)
         (Bound : Vector.t A n) :=
    { bindex : A;
      indexb : IndexBound bindex Bound
    }.

  Import Vectors.VectorDef.VectorNotations.

  Definition BoundedIndex_nil
             (AnyT : Type)
             (idx : BoundedIndex 0 [])
  : AnyT.
  Proof.
    destruct idx as [idx [n] nth_n].
    inversion n.
  Qed.

  Global Arguments indexb {n Bound} b.
  Global Arguments bindex {n Bound} b.
  Global Arguments BoundedIndex {n} Bound%vector_scope.

  Lemma indexb_ibound_eq :
    forall n Bound (bidx bidx' : BoundedIndex (n := n) Bound),
      ibound (indexb bidx) = ibound (indexb bidx') ->
      bindex bidx = bindex bidx'.
  Proof.
    intros; induction Bound; simpl in *.
    - apply BoundedIndex_nil; auto.
    - destruct bidx as [bindex [n'] nth_n];
      destruct bidx' as [bindex' [n''] nth_n']; simpl in *; subst.
      congruence.
  Qed.

  Section Bounded_Index_Dec_Eq.
  (* If equality on A is decideable, so is equality
     on bounded indices in A. *)
    Variable A_eq_dec :
      forall a a' : A, {a = a'} + {a <> a'}.

    Require Import Coq.Logic.Eqdep_dec.

    Program Definition Opt_A_eq_dec (a a' : option A) :
      {a = a'} + {a <> a'} :=
      match a as a, a' as a' return {a = a'} + {a <> a'} with
          | Some a, Some a' => if A_eq_dec a a' then left _ else right _
          | None, None => left _
          | _, _ => right _
      end.

    (*Definition K_Opt_A :
      forall (a : option A) (P: a = a -> Prop),
          P (refl_equal a) -> forall p : a = a, P p :=
      K_dec_set Opt_A_eq_dec.

    Definition UIP_Opt_A  :
      forall (a : option A) (Q : option A -> Type)
             (x : Q a) (h : a = a),
        x = eq_rect a Q x a h.
      intros; eapply K_Opt_A with (p := h); reflexivity.
    Defined.

    Lemma eq_proofs_unicity_Opt_A
      : forall (a a' : option A) (p1 p2 : a = a'), p1 = p2.
      apply eq_proofs_unicity; intros.
      destruct (Opt_A_eq_dec x y); auto.
    Defined. *)

    Corollary idx_ibound_eq
    : forall n Bound (idx idx' : BoundedIndex (n := n) Bound),
        ibound (indexb idx) = ibound (indexb idx') ->
        idx = idx'.
    Proof.
      intros; generalize (indexb_ibound_eq _ _ idx idx' H);
      destruct idx as [idx [n'] In_n'];
        destruct idx' as [idx' [n''] In_n'' ]; intros;
        simpl in *; subst; f_equal.
    Qed.

    Corollary idx_ibound_neq
    : forall n Bound (idx idx' : BoundedIndex (n := n) Bound),
        ibound (indexb idx) <> ibound (indexb idx') ->
        idx <> idx'.
    Proof.
      intros; destruct idx as [idx [n'] In_n' ];
        destruct idx' as [idx' [n''] In_n'' ]; intros;
        simpl in *; subst.
      unfold not; intros; apply H; injection H0; auto.
    Qed.

    Definition fin_eq_dec {m} (n n' : Fin.t m)
      : {n = n'} + {n <> n'}.
    Proof.
      eapply Fin.rect2 with (P := fun m (n n' : Fin.t m)
                                  => {n = n'} + {n <> n'}).
      - left; reflexivity.
      - right; intro; discriminate.
      - right; intro; discriminate.
      - destruct 1; subst.
        + left; reflexivity.
        + right; intro H; apply Fin.FS_inj in H; eauto.
    Defined.

    Corollary BoundedIndex_eq_dec n Bound :
      forall idx idx' : BoundedIndex (n := n) Bound,
        {idx = idx'} + {idx <> idx'}.
    Proof.
      intros; destruct (fin_eq_dec (ibound (indexb idx)) (ibound (indexb idx'))).
      - left; auto using idx_ibound_eq.
      - right; auto using idx_ibound_neq.
    Defined.

    Fixpoint fin_beq {m n} (p : Fin.t m) (q : Fin.t n) :=
      match p, q with
      | @Fin.F1 m', @Fin.F1 n' => EqNat.beq_nat m' n'
      | Fin.FS _, Fin.F1 => false
      | Fin.F1,  Fin.FS _ => false
      | Fin.FS p', Fin.FS q' => fin_beq p' q'
      end.

    Lemma fin_beq_dec {n}
      : forall (p : Fin.t n) (q : Fin.t n),
        fin_beq p q = true <-> p = q.
    Proof.
      intros; pattern n, p, q; eapply Fin.rect2; simpl;
        intuition; try (congruence || discriminate).
      - symmetry; eapply beq_nat_refl.
      - eauto using Fin.FS_inj.
    Qed.

    Corollary fin_beq_neq_dec {n}
      : forall (p : Fin.t n) (q : Fin.t n),
        fin_beq p q = false <-> p <> q.
    Proof.
      intros; setoid_rewrite <- fin_beq_dec; split;
        intros; destruct (fin_beq p q); congruence.
    Qed.

    Corollary fin_beq_refl {m}
      : forall (p : Fin.t m),
        fin_beq p p = true.
    Proof.
      intros; eapply fin_beq_dec; reflexivity.
    Qed.

    Corollary fin_beq_sym {n}
      : forall (p : Fin.t n) (q : Fin.t n),
        fin_beq p q = fin_beq q p.
    Proof.
      intros; case_eq (fin_beq p q); case_eq (fin_beq q p);
        intros; eauto.
      - apply fin_beq_neq_dec in H; apply fin_beq_dec in H0;
          congruence.
      - apply fin_beq_neq_dec in H0; apply fin_beq_dec in H;
          congruence.
    Qed.

    Corollary fin_beq_trans {n}
      : forall (p q r : Fin.t n),
        fin_beq p q = true ->
        fin_beq q r = true ->
        fin_beq p q = true.
    Proof.
      repeat setoid_rewrite fin_beq_dec; congruence.
    Qed.

    Definition BoundedIndex_beq {n Bound}
               (idx idx' : BoundedIndex (n := n) Bound)
      : bool := fin_beq (ibound (indexb idx)) (ibound (indexb idx')).

    Corollary BoundedIndex_beq_dec {n Bound}
      : forall (idx idx' : BoundedIndex (n := n) Bound),
        BoundedIndex_beq idx idx' = true <-> idx = idx'.
    Proof.
      intros; setoid_rewrite fin_beq_dec; split; intros.
      - eauto using idx_ibound_eq.
      - subst; reflexivity.
    Qed.

    Corollary BoundedIndex_beq_neq_dec {n Bound}
      : forall (idx idx' : BoundedIndex (n := n) Bound),
        BoundedIndex_beq idx idx' = false <-> idx <> idx'.
    Proof.
      intros; setoid_rewrite <- BoundedIndex_beq_dec; split;
        intros; destruct (BoundedIndex_beq idx idx'); congruence.
    Qed.

    Corollary BoundedIndex_beq_refl {m Bound}
      : forall (p : BoundedIndex (n := m) Bound),
        BoundedIndex_beq p p = true.
    Proof.
      intros; eapply BoundedIndex_beq_dec; reflexivity.
    Qed.

    Corollary BoundedIndex_beq_sym {n Bound}
      : forall (p q : BoundedIndex (n := n) Bound),
        BoundedIndex_beq p q = BoundedIndex_beq q p.
    Proof.
      intros; case_eq (BoundedIndex_beq p q); case_eq (BoundedIndex_beq q p);
        intros; eauto.
      - apply BoundedIndex_beq_neq_dec in H; apply BoundedIndex_beq_dec in H0;
          congruence.
      - apply BoundedIndex_beq_neq_dec in H0; apply BoundedIndex_beq_dec in H;
          congruence.
    Qed.

    Corollary BoundedIndex_beq_trans {n Bound}
      : forall (p q r : BoundedIndex (n := n) Bound),
        BoundedIndex_beq p q = true ->
        BoundedIndex_beq q r = true ->
        BoundedIndex_beq p q = true.
    Proof.
      repeat setoid_rewrite BoundedIndex_beq_dec; congruence.
    Qed.

  End Bounded_Index_Dec_Eq.

End BoundedIndex.

Section BoundedIndexExtensions.
  (* Helper functions for extending the set of Bounded Indexes *)

  Lemma IndexBound_AppendR {A n m}
  : forall (Bound : Vector.t A n)
           (Bound' : Vector.t A m)
           (idx : Fin.t n),
    Vector.nth (Vector.append Bound Bound') (Fin.L m idx) =
    Vector.nth Bound idx.
  Proof.
    induction Bound; intros.
    - inversion idx.
    - revert Bound IHBound; pattern n, idx; apply Fin.caseS;
        simpl; intros; eauto.
  Qed.

  Definition BoundedIndex_injR {A n Bound m Bound'}
             (idx : BoundedIndex (A := A) (n := n) Bound)
    : BoundedIndex (n := n + m) (Vector.append Bound Bound').
    refine {| bindex := bindex idx;
              indexb :=  {| ibound := Fin.L _ (ibound (indexb idx))|}
           |}.
    (* Abstract Proof Term. *)
    abstract (rewrite <- (boundi (IndexBound := indexb idx));
              eapply IndexBound_AppendR).
  Defined.

  Lemma IndexBound_AppendL {A n m}
    : forall (Bound : Vector.t A n)
             (Bound' : Vector.t A m)
             (idx : Fin.t n),
      Vector.nth (Vector.append Bound' Bound) (Fin.R m idx) =
      Vector.nth Bound idx.
  Proof.
    induction Bound'; intros.
    - reflexivity.
    - simpl; eauto.
  Qed.

  Definition BoundedIndex_injL {A n Bound m Bound'}
             (idx : BoundedIndex (A := A) (n := n) Bound)
    : BoundedIndex (n := m + n) (Vector.append Bound' Bound).
    refine {| bindex := bindex idx;
              indexb :=  {| ibound := Fin.R _ (ibound (indexb idx))|}
           |}.
    (* Abstract Proof Term. *)
    abstract (rewrite <- (boundi (IndexBound := indexb idx));
              eapply IndexBound_AppendL).
  Defined.

End BoundedIndexExtensions.

Ltac Build_nth_IndexBound n A a As As' m :=
  match n with
  | S ?n' =>
    (let check := constr:(eq_refl : a = Vector.hd As') in (* Check if the terms match *)
     exact (@Build_IndexBound A _ a As (Fin.R m (@Fin.F1 n'))
                              (@eq_refl A a)))
      ||
      let As'' := eval simpl in (Vector.tl As') in
          Build_nth_IndexBound n' A a As As'' (S m)
  end.

Hint Extern 0 (@IndexBound ?A ?n ?a ?As) =>
let n' := eval compute in n in
    Build_nth_IndexBound n' A a As As 0 : typeclass_instances.

(*
Import Vectors.VectorDef.VectorNotations.

Fixpoint foo (n : nat) : Vector.t nat (S n) :=
  match n with
  | 0 => 0 :: []
  | S n' => (S n') * (S n') * (S n') :: (foo n')
  end.

Definition foo100 := Eval simpl in foo 20.

Definition baz := Eval simpl in 3 * 3 * 3.

Definition bar : IndexBound baz foo100.
Proof.
  Unset Ltac Debug.
  Time eauto with typeclass_instances.
Defined. *)

Definition BoundedString {n} := @BoundedIndex string n.
Definition BoundedString_eq_dec
           {n}
           {Bound}
           (bidx bidx' : BoundedString Bound)
: {bidx = bidx'} + {bidx <> bidx'} :=
  BoundedIndex_eq_dec n Bound bidx bidx'.

Notation "x ++ y" := (Vector.append x y) : vector_scope.

Notation "`` A" :=
  ({| bindex := A%string |}) (at level 0, format "`` A").

Arguments BoundedString {n} _%vector_scope.

Section ithIndexBound.

  Require Import ADT.ilist.
  Import Vectors.VectorDef.VectorNotations.

  (* Given a bounded index [BoundedIndex Bound], we can wrap
     various lookup functions over lists indexed over [Bound].
   *)

  Context {A : Type}.

  Lemma None_neq_Some
  : forall (AnyT AnyT' : Type) (a : AnyT),
      None = Some a -> AnyT'.
  Proof.
    intros; discriminate.
  Qed.

  Definition nth_Bounded
             {n}
             {Bound : Vector.t A n}
             (idx : BoundedIndex Bound)
    : A := Vector.nth Bound (ibound (indexb idx)).

  Definition ith_Bounded
             {m}
             {B : A -> Type}
             {Bound : Vector.t A m}
             (il : ilist Bound)
             (idx : BoundedIndex Bound)
    : B (Vector.nth Bound (ibound (indexb idx))) :=
    ith il (ibound (indexb idx)) .

  Lemma ith_Bounded_imap
        {m}
        {B B' : A -> Type}
  : forall (f : forall idx, B idx -> B' idx)
           (Bound : Vector.t A m)
           (idx : BoundedIndex Bound)
           (il : ilist Bound),
      f _ (ith_Bounded il idx) =
      ith_Bounded (imap _ _ f Bound il) idx.
  Proof.
    unfold ith_Bounded; simpl; intros; eapply ith_imap.
  Qed.

  Definition replace_BoundedIndex
             {m}
           {B : A -> Type}
           {Bound : Vector.t A m}
           (il : ilist Bound)
           (idx : BoundedIndex Bound)
           (new_b : B (nth_Bounded idx))
  : ilist Bound :=
    replace_Index Bound il (ibound (indexb idx)) new_b.

  Lemma ith_replace_BoundIndex_neq
        {m}
        {B : A -> Type}
  : forall
      (Bound : Vector.t A m)
      (il : ilist Bound)
      (idx idx' : BoundedIndex Bound)
      (new_b : B (nth_Bounded idx')),
      idx <> idx'
      -> ith_Bounded (replace_BoundedIndex il idx' new_b) idx =
         ith_Bounded il idx.
  Proof.
    unfold ith_Bounded; simpl; intros; eapply ith_replace_Index_neq.
    eauto using idx_ibound_eq.
  Qed.

  Lemma ith_replace_BoundIndex_eq
        {m}
        {B : A -> Type}
  : forall
      (Bound : Vector.t A m)
      (il : ilist Bound)
      (idx : BoundedIndex Bound)
      (new_b : B (nth_Bounded idx)),
      ith_Bounded (replace_BoundedIndex il idx new_b) idx = new_b.
  Proof.
    unfold ith_Bounded; simpl; intros; eapply ith_replace_Index_eq.
  Qed.

  Lemma ilist_eq {m} {B : A -> Type}
  : forall (Bound : Vector.t A m)
           (il il' : ilist (B := B) Bound),
      (forall idx, ith_Bounded il idx = ith_Bounded il' idx) -> il = il'.
  Proof.
    induction Bound; intros.
    - rewrite (ilist_invert _ il), (ilist_invert _ il'); reflexivity.
    - destruct il; destruct il'; simpl in *.
      f_equal.
      + generalize (H {| bindex := h |});
        unfold ith_Bounded; simpl; auto.
      + apply IHBound; intros.
        exact (H {| bindex := bindex idx;
                   indexb := @Build_IndexBound
                               _
                               (S n)
                               (bindex idx)
                               (h :: Bound)
                               (Fin.FS (ibound (indexb idx)))
                               (@boundi _ _ _ _ (indexb idx))
                |}).
  Qed.

  Global Arguments imap {A B B'} f {n} {As} il.
  Global Arguments BoundedIndex_eq_dec {_ _ _} _ _.

  Lemma imap_replace_BoundedIndex
        {m}
        {B B' : A -> Type}
  : forall (f : forall idx'', B idx'' -> B' idx'')
           (Bound : Vector.t A m)
           (idx : BoundedIndex Bound)
           (il : ilist Bound)
           b,
      imap f (replace_BoundedIndex il idx b) =
      replace_BoundedIndex (imap f il) idx (f _ b).
  Proof.
    intros; apply ilist_eq; intros.
    destruct (BoundedIndex_eq_dec idx idx0); subst;
    rewrite <- ith_Bounded_imap.
    - repeat rewrite ith_replace_BoundIndex_eq; reflexivity.
    - repeat rewrite ith_replace_BoundIndex_neq; auto.
      rewrite <- ith_Bounded_imap; auto.
  Qed.

End ithIndexBound.


Arguments BoundedString [_] _.
